import functools
import logging
import threading
from collections import MutableMapping
from contextlib import contextmanager

from tornado.stack_context import StackContext

from dorthy.dp import Observable


logger = logging.getLogger(__name__)


class RequestContextError(Exception):
    pass


class RequestLocalStore(threading.local):

    def __init__(self):
        super().__init__()
        self.__context = None

    @property
    def active(self):
        return True if self.__context is not None else False

    @property
    def context(self):
        return self.__context

    @context.setter
    def context(self, context):
        assert not self.active
        self.__context = context

    def release(self):
        self.__context = None

_request_store = RequestLocalStore()


class RequestContext(MutableMapping):

    def __init__(self):
        self.__active = False
        self.__store = dict()
        self.__immutable = set()
        self.__observable = Observable()

    def __contains__(self, item):
        return item in self.__store

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, item):
        return self.__store[item]

    def __delitem__(self, key):
        self.__check_immutable(key)
        del self.__store[key]

    def __setitem__(self, key, value):
        self.__check_immutable(key)
        self.__store[key] = value

    def __iter__(self):
        return iter(self.__store)

    @property
    def active(self):
        return self.__active

    @property
    def id(self):
        return id(self)

    def __check_immutable(self, key):
        if key in self.__immutable:
            raise ValueError("The value is immutable and cannot be changed.")

    def immutable(self, key, value):
        """Sets the given value and makes it immutable for the
        lifetime of the request context.  If someone tries to change
        the value then a ValueError is raised.
        """
        self.__check_immutable(key)
        self[key] = value
        self.__immutable.add(key)

    def contains_listener(self, listener):
        return listener in self.__observable

    def register_listener(self, listener):
        self.__observable.register(listener)

    def _activate(self):
        _request_store.context = self
        self.__active = True
        self.__observable("activate", self)

    def _release(self):
        if self.__active:
            # fire listener first because listeners may depend on
            # request to perform cleanup work
            self.__observable("deactivate", self)
            _request_store.release()
            self.__active = False


class RequestContextManager(object):

    def __init__(self, request_context=None):
        # None check required because request_context acts like a dict
        self.__request_context = request_context if request_context is not None else RequestContext()
        self.__nesting = 0

    @contextmanager
    def context_manager(self):
        if self.__nesting == 0:
            self.__request_context._activate()
        self.__nesting += 1
        try:
            yield
        except Exception as e:
            # on exception release immediately
            self.__nesting = 0
            self.__request_context._release()
            logger.warn("Encounter an exception in request context manager.")
            raise e
        else:
            if self.__nesting <= 0:
                # should not happen but handle for safety
                self.__nesting = 0
                self.__request_context._release()
                logger.warn("Request context release called without active context.")
            else:
                self.__nesting -= 1
                if self.__nesting == 0:
                    self.__request_context._release()

    @staticmethod
    def active():
        return _request_store.active

    @staticmethod
    def get_context():
        if not _request_store.active:
            raise RequestContextError("No request context found for executing stack.")
        return _request_store.context


class ProxyHandlerMetaClass(type):

    def __init__(cls, name, base, dct):
        super().__init__(name, base, dct)
        setattr(cls, "__getattribute__", ProxyHandlerMetaClass.__getattribute)

    @staticmethod
    def request_proxy(method):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            request_context = None
            if hasattr(method.__self__, "_request_context"):
                request_context = method.__self__._request_context
            with StackContext(RequestContextManager(request_context).context_manager):
                return method(*args, **kwargs)
        return wrapper

    @staticmethod
    def __getattribute(obj, name):
        attr = object.__getattribute__(obj, name)
        if name in object.__getattribute__(obj, "PROXY_METHODS"):
            if hasattr(obj, "STORE_CONTEXT_ON_OBJECT") and not hasattr(obj, "_request_context"):
                obj._request_context = RequestContext()
            attr = ProxyHandlerMetaClass.request_proxy(attr)
        return attr


class SocketHandlerProxyMixin(metaclass=ProxyHandlerMetaClass):

    STORE_CONTEXT_ON_OBJECT = True
    PROXY_METHODS = ("on_open", "on_message", "on_close")


class WebRequestHandlerProxyMixin(metaclass=ProxyHandlerMetaClass):

    PROXY_METHODS = "_execute"



import glob
import logging
import os
import yaml

logger = logging.getLogger(__name__)

config = None


class Properties(object):
    """Provides a property / attribute representation of a hierarchical
    dictionary of values"""

    def __init__(self, props):
        for name, value in props.items():
            if isinstance(value, dict):
                setattr(self, name, Properties(value))
            else:
                setattr(self, name, value)

    def __setattr__(self, name, value):
        if not hasattr(self, name):
            object.__setattr__(self, name, value)
        else:
            raise AttributeError("Cannot set values on properties objects")

    def __delattr__(self, name):
        raise AttributeError("Cannot delete values on properties objects")

    def __call__(self):
        """Calling an instance of a Property returns a dict representation
        of the property tree"""
        d = dict()
        for name, value in self.__dict__.items():
            if isinstance(value, Properties):
                d[name] = value()
            else:
                d[name] = value
        return d

    def __contains__(self, item):
        keys = item.split(".")
        prop_obj = self
        for key in keys:
            try:
                prop_obj = getattr(prop_obj, key)
            except AttributeError:
                return False
        return True

    def _asdict(self):
        return self.__call__()

    def enabled(self, key):
        try:
            return True if getattr(self, key) else False
        except AttributeError:
            return False

    def get(self, key, default=None, split=False):
        try:
            value = getattr(self, key)
        except AttributeError:
            return default
        else:
            if split:
                value = value.split(",")
                # remove the empty string from the end
                if not value[-1]:
                    value = value[:-1]
            return value


def _load_path(pathname):
    """Load yaml conf files for the given path into
    the module dictory configs"""
    props = {}
    path = os.path.join(pathname, "*.yml")
    for confpath in glob.iglob(path):
        with open(confpath) as f:
            key = os.path.basename(confpath)[:-4]
            conf = yaml.load(f)
            props[key] = conf

    global config
    if config is None:
        config = Properties(props)
    else:
        # reload config with overrides
        d = config._asdict()
        for name, value in props.items():
            d[name] = value
        config = Properties(d)


# load default path from working directory
_conf_dir = os.path.join(os.getcwd(), "conf")
_load_path(_conf_dir)

# load overrides
if os.getenv("CONF_OVERRIDE"):
    _load_path(os.path.join(_conf_dir, os.getenv("CONF_OVERRIDE")))


import base64
import time
import logging
import zlib

from functools import partial

import raven
from raven.base import Client
from raven.utils import get_auth_header

from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from raven.contrib.tornado import SentryMixin

from dorthy.json import jsonify
from dorthy.security import SecurityManager

logger = logging.getLogger(__name__)


class TornadoSentryClient(Client):
    """Initial inspiration from raven.contrib.tornado
    but the raven version is broken for handling errors out
    of the AsyncHTTPClient and does not support disabling
    SSL server cert validation.
    """

    def __init__(self, dsn=None, host_id=None, validate_cert=False, **options):
        super().__init__(dsn=dsn, **options)
        self.__host_id = host_id
        self.__validate_cert = validate_cert

    @property
    def host_id(self):
        return self.__host_id if self.__host_id else ""

    def capture(self, *args, **kwargs):
        """
        Takes the same arguments as the super function in :py:class:`Client`
        and extracts the keyword argument callback which will be called on
        asynchronous sending of the request

        :return: a 32-length string identifying this event and checksum
        """
        if not self.is_enabled():
            return

        data = self.build_msg(*args, **kwargs)

        self.send(callback=kwargs.get('callback', None), **data)

        return data['event_id'],

    def encode(self, data):
        """
        Serializes ``data`` into a raw string.
        """
        return base64.b64encode(zlib.compress(jsonify(data).encode('utf8')))

    def send(self, auth_header=None, callback=None, **data):
        """
        Serializes the message and passes the payload onto ``send_encoded``.
        """

        try:
            message = self.encode(data)
        except TypeError:
            data = {"message": "Failed to serialize data for message. See server for details. Host ID: {}".
                    format(self.host_id)}
            message = self.encode(data)

        return self.send_encoded(message, auth_header=auth_header, callback=callback)

    def send_encoded(self, message, auth_header=None, **kwargs):
        """
        Given an already serialized message, signs the message and passes the
        payload off to ``send_remote`` for each server specified in the servers
        configuration.

        callback can be specified as a keyword argument
        """
        if not auth_header:
            timestamp = time.time()
            auth_header = get_auth_header(
                protocol=self.protocol_version,
                timestamp=timestamp,
                client='raven-python/%s' % (raven.VERSION,),
                api_key=self.public_key,
                api_secret=self.secret_key,
            )

        for url in self.servers:
            headers = {
                'X-Sentry-Auth': auth_header,
                'Content-Type': 'application/octet-stream',
            }

            self.send_remote(
                url=url, data=message, headers=headers,
                callback=kwargs.get('callback', None)
            )

    def send_remote(self, url, data, headers=None, callback=None):

        if not self.state.should_try():
            message = self._get_log_message(data)
            self.error_logger.error(message)
            return

        # TODO: implement retry logic -- the base implementation does
        # not work for asynchronous tornado processing.  It is based on
        # a singleton pattern for the client.
        self._send_remote(url, data, headers=headers, callback=callback)
        self.state.set_success()

    def _send_remote(self, url, data, headers=None, callback=None):

        if headers is None:
            headers = dict()

        # disable cert validation
        request = HTTPRequest(url,
                              method="POST",
                              body=data,
                              headers=headers,
                              validate_cert=self.__validate_cert)

        AsyncHTTPClient().fetch(request,
                                callback=partial(self._handle_response, url, data, callback))

    def _handle_response(self, url, data, callback, response):

        if response.error:
            self.error_logger.error(
                'Unable to reach Sentry log server: %s - Error Code: %s Message: %s',
                url, str(response.code), str(response.error)
            )
            message = self._get_log_message(data)
            self.error_logger.error('Failed to submit message: %r', message)
            return

        # try and process possible callback
        if callback and callable(callback):
            try:
                callback(response)
            except:
                self.error_logger.exception("Failed to process callback.")


class TornadoSentryMixin(SentryMixin):
    """
    A Sentry Mixin class for base tornado
    """

    def get_sentry_user_info(self):
        """
        Data for sentry.interfaces.User

        Default implementation only sends `is_authenticated` by checking if
        `tornado.web.RequestHandler.get_current_user` tests postitively for on
        Truth calue testing
        """
        auth = SecurityManager().get_authentication() if SecurityManager().authenticated() else None
        return {
            'sentry.interfaces.User': {
                'is_authenticated': True if auth else False,
                'id': auth.principal.uid if auth else None,
                'principal': auth.principal.name if auth else None
            }
        }


class SockJSSentryMixin(TornadoSentryMixin):
    """
    A SentryMixin class specifically for use with sockjs handler
    """
    
    def get_sentry_client(self):
        """
        Returns the sentry client configured in the application.
        Need to set sentry_client var during server setup
        i.e.:
        application.sentry_client = AsyncSentryClient(config.app.sentry.server)
        application.SocketRouter.sentry_client = application.sentry_client
        """
        return getattr(self.session.server, 'sentry_client', None)
    
    def get_sentry_data_from_request(self):
        """
        Extracts the data required for 'sentry.interfaces.Http' from the
        SockJS ConnectionInfo object

        :param return: A dictionary.
        """
        return {
            'sentry.interfaces.Http': {
                'url': self.request.path,
                'method': 'SOCKJS',
                'data': self.request.arguments,
                'cookies': self.request.cookies,
                'headers': dict(self.request.headers),
                'env': {
                    'REMOTE_ADDR': self.request.ip,
                }
            }
        }
    
    def get_default_context(self):
        data = {}

        # Update request data
        data.update(self.get_sentry_data_from_request())

        # update user data
        data.update(self.get_sentry_user_info())

        # Update extra data
        data.update(self.get_sentry_extra_info())

        return data
    
    def _capture(self, call_name, data=None, **kwargs):
        client = self.get_sentry_client()
        if not client:
            return
        
        if data is None:
            data = self.get_default_context()
        else:
            default_context = self.get_default_context()
            if isinstance(data, dict):
                default_context.update(data)
            else:
                default_context['extra']['extra_data'] = data
            data = default_context
            
        logger.debug('Sending to Sentry via method {}: {}'.format(call_name, data))

        return getattr(client, call_name)(data=data, **kwargs)

    def captureException(self, exc_info=None, **kwargs):
        return self._capture('captureException', exc_info=exc_info, **kwargs)

    def captureMessage(self, message, **kwargs):
        return self._capture('captureMessage', message=message, **kwargs)